from dataclasses import dataclass
import carla
from carla import Location, Rotation, Transform
import math
import numpy as np
from collections import deque
import socket
import random
import time
from typing import Tuple, List
from invertedai.simulators.data.static_carla import (
    MAP_CENTERS,
    DEMO_LOCATIONS,
    NPC_BPS,
    EGO_FLAG_COLOR,
    NPC_FLAG_COLOR,
    RS,
    cord,
)


@dataclass
class CarlaSimulationConfig:
    npc_bps: Tuple[str] = NPC_BPS
    roi_center: cord = cord(x=0, y=0)  # region of interest center
    map_name: str = "Town03"
    scene_name: str = "CARLA:Town03:Roundabout"
    fps: int = 10  # Should not be compatible with invertedAI fps
    traffic_count: int = 20
    episode_length: int = 20  # In Seconds
    entrance_interval: int = 2  # In Seconds
    follow_ego: bool = False
    slack: int = 3
    ego_bp: str = "vehicle.tesla.model3"
    seed: float = time.time()
    flag_npcs: bool = True
    flag_ego: bool = True
    ego_autopilot: bool = True
    npcs_autopilot: bool = False
    populate_npcs: bool = True
    npc_population_interval: int = 1  # In Seconds
    non_roi_npc_mode: str = (
        "carla_handoff"  # ["no_non_roi_npc", "spawn_at_entrance", "carla_handoff"]
    )
    max_cars_in_map: int = 100
    proximity_threshold = None


class Car:
    def __init__(self, actor: carla.Actor, speed: float = 0.0) -> None:
        self.actor = actor
        self.recurrent_state = RS
        self._dimension = None
        self._states = deque(maxlen=10)
        self.speed = speed

    def update_dimension(self):
        self._dimension = self._get_actor_dimensions()

    def update_speed(self):
        v = self.actor.get_velocity()
        vs = np.sqrt(v.x**2 + v.y**2)
        self.speed = vs

    @property
    def dims(self):
        return self._dimension

    def get_state(self, from_carla=False):
        self.transform, state = self._get_actor_state(from_carla)
        self._states.append(state)
        return dict(
            transform=self.transform,
            states=list(self._states),
            recurrent_state=self.recurrent_state,
        )

    def set_state(self, state=None, recurrent_state=None):
        self.recurrent_state = recurrent_state
        if state is not None:
            # NOTE: state is of size 4 : [x, y, angle, speed]
            loc = carla.Location(state[0], state[1], self.transform.location.z)
            rot = carla.Rotation(
                yaw=np.degrees(state[2]),
                pitch=self.transform.rotation.pitch,
                roll=self.transform.rotation.roll,
            )
            next_transform = carla.Transform(loc, rot)
            self.actor.set_transform(next_transform)
            self.speed = state[3]

    def _get_actor_dimensions(self):
        bb = self.actor.bounding_box.extent
        length = max(
            2 * bb.x, 1.0
        )  # provide minimum value since CARLA returns 0 for some agents
        width = max(2 * bb.y, 0.2)
        physics_control = self.actor.get_physics_control()
        # Wheel position is in centimeter: https://github.com/carla-simulator/carla/issues/2153
        rear_left_wheel_position = physics_control.wheels[2].position / 100
        rear_right_wheel_position = physics_control.wheels[3].position / 100
        real_mid_position = 0.5 * (rear_left_wheel_position + rear_right_wheel_position)
        actor_geo_center = self.actor.get_location()
        lr = actor_geo_center.distance(real_mid_position)
        return (length, width, lr)

    def _get_actor_state(self, from_carla=False):
        t = self.actor.get_transform()
        loc, rot = t.location, t.rotation
        xs = loc.x
        ys = loc.y
        psis = np.radians(rot.yaw)
        if from_carla:
            v = self.actor.get_velocity()
            vs = np.sqrt(v.x**2 + v.y**2)
        else:
            vs = self.speed
        return t, (xs, ys, psis, vs)


class CarlaEnv:
    def __init__(
        self,
        cfg: CarlaSimulationConfig,
        ego_spawn_point=None,
        initial_states=None,
        npc_entrance_spawn_points=None,
        spectator_transform=None,
    ) -> None:
        self.rng = random.Random(cfg.seed)
        self.cfg = cfg
        centers = MAP_CENTERS[cfg.scene_name]
        self.cfg.roi_center = cord(x=centers[0], y=centers[1])
        self.cfg.map_name = cfg.scene_name.split(":")[1]
        world_settings = carla.WorldSettings(
            synchronous_mode=True,
            fixed_delta_seconds=1 / float(cfg.fps),
        )
        client = carla.Client("localhost", 2000)
        traffic_manager = client.get_trafficmanager(
            get_available_port(subsequent_ports=0)
        )
        world = client.load_world(cfg.map_name)
        self.original_settings = client.get_world().get_settings()
        world.apply_settings(world_settings)
        traffic_manager.set_synchronous_mode(True)
        traffic_manager.set_hybrid_physics_mode(True)
        self.proximity_threshold = cfg.proximity_threshold or (
            50
            if cfg.scene_name not in DEMO_LOCATIONS.keys()
            else DEMO_LOCATIONS[cfg.scene_name]["proximity_threshold"]
        )
        if initial_states is None:
            spawn_points = world.get_map().get_spawn_points()
            npc_roi_spawn_points, initial_speed = self.get_roi_spawn_points(
                spawn_points, speed=np.zeros_like(spawn_points)
            )
        else:
            spawn_points, speed = self._to_transform(initial_states)
            npc_roi_spawn_points, initial_speed = self.get_roi_spawn_points(
                spawn_points, speed
            )
        if (ego_spawn_point is None) or (cfg.scene_name not in DEMO_LOCATIONS.keys()):
            ego_spawn_point, _ = (npc_roi_spawn_points.pop(), initial_speed.pop())
        elif ego_spawn_point == "demo":
            locs = DEMO_LOCATIONS[cfg.scene_name]
            ego_spawn_point = self.rng.choice(locs["spawning_locations"])
        else:
            assert isinstance(
                ego_spawn_point, carla.Transform
            ), "ego_spawn_point must be a Carla.Transform"
        if cfg.non_roi_npc_mode == "spawn_at_entrance":
            self.nroi_npc_mode = 0
            # TODO: use enum to combine self.nroi_npc_mode and cfg.non_roi_npc_mode
            if npc_entrance_spawn_points is None:
                spawn_points = world.get_map().get_spawn_points()
                npc_entrance_spawn_points = self.get_entrance(spawn_points)
            else:
                spawn_points = self._to_transform(npc_roi_spawn_points)
                npc_entrance_spawn_points = self.get_roi_spawn_points(spawn_points)
        elif cfg.non_roi_npc_mode == "carla_handoff":
            self.nroi_npc_mode = 1
            spawn_points = world.get_map().get_spawn_points()
            self.non_roi_spawn_points, _ = self.get_roi_spawn_points(
                spawn_points, roi=False
            )
        else:
            self.nroi_npc_mode = 2
        if spectator_transform is None:
            camera_loc = carla.Location(cfg.roi_center.x, cfg.roi_center.y, z=100)
            camera_rot = carla.Rotation(pitch=-90, yaw=90, roll=0)
            spectator_transform = carla.Transform(camera_loc, camera_rot)
            self.spectator_mode = "birdview"
        elif spectator_transform == "follow_ego":
            spectator_transform = carla.Transform(
                ego_spawn_point.transform(carla.Location(x=-6, z=2.5)),
                ego_spawn_point.rotation,
            )
            self.spectator_mode = "follow_ego"
        else:
            assert isinstance(
                spectator_transform, carla.Transform
            ), "spectator_transform must be a Carla.Transform"
            self.spectator_mode = "user_defined"

        self.spectator = world.get_spectator()
        self.spectator.set_transform(spectator_transform)
        self.world = world
        self.client = client
        self.traffic_manager = traffic_manager
        self.roi_spawn_points = npc_roi_spawn_points
        self.initial_speed = initial_speed
        self.entrance_spawn_points = npc_entrance_spawn_points
        self.ego_spawn_point = ego_spawn_point
        self.populate_step = self.cfg.fps * self.cfg.npc_population_interval
        self.npcs = []
        self.non_roi_npcs = []

    def _initialize(self):
        # Keep the order of first spawining ego then NPCs
        # to avoid spawning npc in ego location
        self.ego = self._spawn_npcs(
            [self.ego_spawn_point],
            [0],
            [self.cfg.ego_bp],
        ).pop()
        if len(self.roi_spawn_points) < self.cfg.traffic_count:
            print("Number of roi_spawn_points is less than traffic_count")
            # TODO: Add logger
        num_npcs = min(len(self.roi_spawn_points), self.cfg.traffic_count)
        self.npcs.extend(
            self._spawn_npcs(
                self.roi_spawn_points[:num_npcs],
                self.initial_speed,
                self.cfg.npc_bps,
            )
        )
        if self.cfg.non_roi_npc_mode == "carla_handoff":
            num_npcs = min(len(self.non_roi_spawn_points), self.cfg.max_cars_in_map)
            self.non_roi_npcs.extend(
                self._spawn_npcs(
                    self.non_roi_spawn_points[:num_npcs],
                    [0 for _ in range(num_npcs)],
                    self.cfg.npc_bps,
                )
            )
        self.set_npc_autopilot(self.npcs, True)
        for _ in range(10):  # Ticking the world to place cars on the ground
            self.world.tick()
        for npc in self.npcs:
            npc.update_dimension()
        self.ego.update_dimension()
        self.set_npc_autopilot(self.npcs, self.cfg.npcs_autopilot)
        self.set_ego_autopilot(self.cfg.ego_autopilot)
        if self.cfg.non_roi_npc_mode == "carla_handoff":
            self.set_npc_autopilot(self.non_roi_npcs, True)
        self.step_counter = 0

    def reset(self, include_ego=True):
        try:
            self.destroy()
        except:
            pass
        self._initialize()
        return self.get_obs(include_ego=include_ego)

    def step(self, ego="autopilot", npcs=None, time_step=0, include_ego=True):
        self.step_counter += 1
        self._set_state_and_filter_npcs(ego, npcs, time_step, include_ego)
        if self.cfg.flag_ego:
            self._flag_npc([self.ego], EGO_FLAG_COLOR)
        if self.cfg.flag_npcs:
            self._flag_npc(self.npcs, NPC_FLAG_COLOR)
        if self.spectator_mode == "follow_ego":
            ego_transform = self.ego.transform
            spectator_transform = carla.Transform(
                ego_transform.transform(carla.Location(x=-6, z=2.5)),
                ego_transform.rotation,
            )
            self.spectator.set_transform(spectator_transform)

        self.world.tick()
        return self.get_obs()

    def destroy(self, npcs=True, ego=True, world=True):
        if npcs:
            self._destory_npcs(self.npcs)
            self._destory_npcs(self.non_roi_npcs)
            self.npcs = []
            self.non_roi_npcs = []
        if ego:
            self._destory_npcs([self.ego])
            self.ego = None
        if world:
            self.client.get_world().apply_settings(self.original_settings)
            self.traffic_manager.set_synchronous_mode(False)

    def get_obs(self, obs_len=1, include_ego=True, warmup=False):
        if self.cfg.non_roi_npc_mode == "spawn_at_entrance":
            if len(self.non_roi_npcs) > 0:
                for npc in self.non_roi_npcs:
                    npc.update_dimension()
                self.npcs.extend(self.non_roi_npcs)
                self.non_roi_npcs = []
                self.set_npc_autopilot(self.npcs, self.cfg.npcs_autopilot)
        states = []
        rec_state = []
        dims = []
        for npc in self.npcs:
            obs = npc.get_state(from_carla=warmup)
            states.append(obs["states"][-obs_len:])
            rec_state.append(obs["recurrent_state"])
            dims.append(npc.dims)
        if include_ego:
            obs = self.ego.get_state(from_carla=True)
            states.append(obs["states"][-obs_len:])
            rec_state.append(obs["recurrent_state"])
            dims.append(self.ego.dims)
        return states, rec_state, dims

    def get_infractions(self):
        pass

    def get_reward(self):
        pass

    def is_done(self):
        pass

    def seed(self, seed=None):
        pass

    def render(self, mode="human"):
        pass

    @staticmethod
    def set_npc_autopilot(npcs, on=True):
        for npc in npcs:
            try:
                npc.actor.set_autopilot(on)
                npc.actor.set_simulate_physics(on)
            except:
                print("Unable to set autopilot")
                # TODO: add logger

    def set_ego_autopilot(self, on=True):
        try:
            self.ego.actor.set_autopilot(on)
        except:
            print("Unable to set autopilot")
            # TODO: add logger

    @classmethod
    def from_preset_data(
        cls,
        ego_spawn_point=None,
        initial_states=None,
        npc_entrance_spawn_points=None,
        spectator_transform=None,
    ):
        cfg = CarlaSimulationConfig()
        return cls(
            cfg,
            ego_spawn_point,
            initial_states,
            npc_entrance_spawn_points,
            spectator_transform,
        )

    def _destory_npcs(self, npcs: List):
        for npc in npcs:
            try:
                npc.actor.set_autopilot(False)
            except:
                print("Unable to set autopilot")
                # TODO: add logger
            npc.actor.destroy()

    def _spawn_npcs(self, spawn_points, speeds, bps):
        npcs = []
        for spawn_point, speed in zip(spawn_points, speeds):
            blueprint = self.world.get_blueprint_library().find(self.rng.choice(bps))
            # ego_spawn_point = self.roi_spawn_points[i]
            actor = self.world.try_spawn_actor(blueprint, spawn_point)
            if actor is None:
                print(f"Cannot spawn NPC at:{str(spawn_point)}")
            else:
                npc = Car(actor, speed)
                npcs.append(npc)
        return npcs

    def _flag_npc(self, actors, color):
        for actor in actors:
            loc = actor.actor.get_location()
            loc.z += 3
            self.world.debug.draw_point(
                location=loc,
                size=0.1,
                color=color,
                life_time=2 / self.cfg.fps,
            )

    def _set_state_and_filter_npcs(
        self, ego="autopilot", npcs=None, time_step=0, include_ego=True
    ):
        if npcs is not None:
            states = npcs["states"]
            recurrent_states = npcs["recurrent_states"]
            id = -1  # In case all NPCs vanish!
            for id, npc in enumerate(self.npcs):
                # NOTE: states is of size (batch_size x actor x time x state)
                # where state is a list: [x, y, angle, speed]
                rs = None if recurrent_states is None else recurrent_states[0][id]
                npc.set_state(states[0][id][time_step], rs)

            if include_ego:
                rs = None if recurrent_states is None else recurrent_states[0][id + 1]
                self.ego.set_state(recurrent_state=rs)
        exit_npcs = []
        remaining_npcs = []
        for npc in self.npcs:
            actor_geo_center = npc.get_state()["transform"].location
            distance = math.sqrt(
                ((actor_geo_center.x - self.cfg.roi_center.x) ** 2)
                + ((actor_geo_center.y - self.cfg.roi_center.y) ** 2)
            )
            if distance < self.proximity_threshold + self.cfg.slack:
                remaining_npcs.append(npc)
            else:
                exit_npcs.append(npc)
        if self.cfg.non_roi_npc_mode == "carla_handoff":
            for npc in self.non_roi_npcs:
                actor_geo_center = npc.get_state()["transform"].location
                distance = math.sqrt(
                    ((actor_geo_center.x - self.cfg.roi_center.x) ** 2)
                    + ((actor_geo_center.y - self.cfg.roi_center.y) ** 2)
                )
                if distance < self.proximity_threshold + self.cfg.slack:
                    npc.update_dimension()
                    self.set_npc_autopilot([npc], on=self.cfg.npcs_autopilot)
                    remaining_npcs.append(npc)
                    npc.update_speed()
                else:
                    exit_npcs.append(npc)
            self.non_roi_npcs = exit_npcs
            self.set_npc_autopilot(self.non_roi_npcs, True)
            exit_npcs = []
        elif self.cfg.non_roi_npc_mode == "spawn_at_entrance":
            if not (self.step_counter % self.populate_step):
                self.non_roi_npcs = self._spawn_npcs(
                    self.entrance_spawn_points,
                    (3 * np.ones_like(self.entrance_spawn_points)).tolist(),
                    self.cfg.npc_bps,
                )

        self._destory_npcs(exit_npcs)
        self.npcs = remaining_npcs

    @staticmethod
    def _to_transform(poses: list) -> Tuple[list, list]:
        t = []
        speed = []
        for pos in poses:
            loc = carla.Location(x=pos[0][0], y=pos[0][1], z=1)
            rot = carla.Rotation(yaw=np.degrees(pos[0][2]))
            t.append(carla.Transform(loc, rot))
            speed.append(pos[0][3])
        return (t, speed)

    def get_entrance(self, spawn_points):
        entrance = []
        for sp in spawn_points:
            distance = math.sqrt(
                ((sp.location.x - self.cfg.roi_center.x) ** 2)
                + ((sp.location.y - self.cfg.roi_center.y) ** 2)
            )
            if (
                self.proximity_threshold - self.cfg.slack
                < distance
                < self.proximity_threshold + self.cfg.slack
            ):
                entrance.append(sp)
        return entrance

    def get_roi_spawn_points(self, spawn_points, speed=None, roi=True):
        roi_spawn_points = []
        initial_speed = []
        for ind, sp in enumerate(spawn_points):
            distance = math.sqrt(
                ((sp.location.x - self.cfg.roi_center.x) ** 2)
                + ((sp.location.y - self.cfg.roi_center.y) ** 2)
            )
            if roi & (distance < self.proximity_threshold):
                roi_spawn_points.append(sp)
                if speed is not None:
                    initial_speed.append(speed[ind])
            elif (not roi) & (distance > self.proximity_threshold):
                roi_spawn_points.append(sp)
        return roi_spawn_points, initial_speed


def get_available_port(subsequent_ports: int = 2) -> int:
    """
    Finds an open port such that the given number of subsequent ports are also available.
    The default number of two ports corresponds to what is required by the CARLA server.

    :param subsequent_ports: How many more subsequent ports need to be free.
    """

    # CARLA server needs three consecutive ports.
    # It is possible for some other process to grab the sockets
    # between finding them here and starting the server,
    # but it's generally unlikely.
    limit = 1000
    for attempt in range(limit):
        first = socket.socket()
        subsequent = [socket.socket() for i in range(subsequent_ports)]
        try:
            first.bind(("", 0))
            port = first.getsockname()[1]
            for i in range(len(subsequent)):
                subsequent[i].bind(("", port + i + 1))
            return port
        except OSError as e:
            if attempt + 1 == limit:
                raise RuntimeError(
                    "Failed to find required ports in %d attempts" % limit
                ) from e
        finally:
            first.close()
            for s in subsequent:
                s.close()
    assert False  # this line should be unreachable



"""
Python SDK for

Functions
---------
.. autosummary::
   :toctree: generated/
    available_maps
    drive
    get_map
    initialize
"""
from invertedai.error import TryAgain
import math
from typing import List, Optional
import time
import invertedai as iai


TIMEOUT = 10


def available_maps(*args):
    """
    Searching the available maps using the provided keywords as *args

    Parameters
    ----------
    *args: str
        Variable length argument list of keywords.

    Returns
    -------
    out : List[str]
        A list of "availalbe maps" to your account (api-key)

    See Also
    --------
    invertedai.get_map

    Notes
    -----
    Providing more than three keywords is uncessary

    Examples
    --------
    >>> iai.available_maps("carla", "roundabout")
    ["CARLA:Town03:Roundabout"]
    """
    start = time.time()
    timeout = TIMEOUT
    keywords = "+".join(list(args))
    while True:
        try:
            params = {
                "keywords": keywords,
            }
            response = iai.session.request(model="available_maps", params=params)
            return response
        except TryAgain as e:
            if timeout is not None and time.time() > start + timeout:
                raise e
            iai.logger.info(iai.logger.logfmt("Waiting for model to warm up", error=e))


def get_map(location="CARLA:Town03:Roundabout", include_map_source=True) -> dict:
    start = time.time()
    timeout = TIMEOUT

    params = {"location": location, "include_map_source": include_map_source}
    while True:
        try:
            response = iai.session.request(model="get_map", params=params)
            return response
        except TryAgain as e:
            if timeout is not None and time.time() > start + timeout:
                raise e
            iai.logger.info(iai.logger.logfmt("Waiting for model to warm up", error=e))


def initialize(
    location="CARLA:Town03:Roundabout",
    agent_count=1,
    batch_size=1,
    min_speed=None,
    max_speed=None,
) -> dict:
    start = time.time()
    timeout = TIMEOUT

    while True:
        try:
            include_recurrent_states = (
                False if location.split(":")[0] == "huawei" else True
            )
            params = {
                "location": location,
                "num_agents_to_spawn": agent_count,
                "num_samples": batch_size,
                "spawn_min_speed": min_speed
                and int(math.ceil(min_speed / 3.6))
                and not include_recurrent_states,
                "spawn_max_speed": max_speed
                and int(math.ceil(max_speed / 3.6))
                and not include_recurrent_states,
                "include_recurrent_states": include_recurrent_states,
            }
            initial_states = iai.session.request(model="initialize", params=params)
            response = {
                "states": initial_states["initial_condition"]["agent_states"],
                "recurrent_states": initial_states["recurrent_states"],
                "attributes": initial_states["initial_condition"]["agent_sizes"],
                "traffic_light_state": initial_states["traffic_light_state"],
                "traffic_states_id": initial_states["traffic_states_id"],
            }
            return response
        except TryAgain as e:
            if timeout is not None and time.time() > start + timeout:
                raise e
            iai.logger.info(iai.logger.logfmt("Waiting for model to warm up", error=e))


def drive(
    location: str = "CARLA:Town03:Roundabout",
    states: dict = {},
    agent_attributes: dict = {},
    recurrent_states: Optional[List] = None,
    get_birdviews: bool = False,
    steps: int = 1,
    get_infractions: bool = False,
    traffic_states_id: str = "000:0",
    exclude_ego_agent: bool = True,
    present_mask: Optional[List] = None,
) -> dict:
    def _tolist(input_data: List):
        if not isinstance(input_data, list):
            return input_data.tolist()
        else:
            return input_data

    recurrent_states = (
        _tolist(recurrent_states) if recurrent_states is not None else None
    )  # Bx(num_predictions)xAxTx2x64

    model_inputs = dict(
        location=location,
        past_observations=dict(
            agent_states=states,
            agent_sizes=agent_attributes,
        ),
        recurrent_states=recurrent_states,
        # Expand from BxA to BxAxT_total for the API interface
        steps=steps,
        get_birdviews=get_birdviews,
        get_infractions=get_infractions,
        traffic_states_id=traffic_states_id,
        exclude_ego_agent=exclude_ego_agent,
        present_mask=present_mask,
    )

    start = time.time()
    timeout = TIMEOUT

    while True:
        try:
            return iai.session.request(model="drive", data=model_inputs)
        except Exception as e:
            # TODO: Add logger
            iai.logger.warning("Retrying")
            if timeout is not None and time.time() > start + timeout:
                raise e



import requests
import json
import re
from typing import Dict, Optional
from requests.auth import AuthBase
import invertedai as iai
from invertedai import error
import logging

TIMEOUT_SECS = 600
MAX_RETRIES = 10


class Session:
    def __init__(self, api_token: str = ""):
        self.session = requests.Session()
        self.session.auth = APITokenAuth(api_token)
        self.session.mount(
            "https://",
            requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES),
        )
        self.session.headers.update(
            {
                "Content-Type": "application/json",
                "Accept-Encoding": "gzip, deflate, br",
                # "Connection": "keep-alive",
            }
        )
        self.base_url = self._get_base_url()

    def add_apikey(self, api_token: str = ""):
        if not iai.dev and not api_token:
            raise error.InvalidAPIKeyError("Empty API key received.")
        self.session.auth = APITokenAuth(api_token)

    def request(
        self, model: str, params: Optional[dict] = None, data: Optional[dict] = None
    ):
        method, relative_path = iai.model_resources[model]
        response = self._request(
            method=method,
            relative_path=relative_path,
            params=params,
            json=data,
        )

        return response

    def _request(
        self,
        method,
        relative_path: str = "",
        params=None,
        headers=None,
        json=None,
        data=None,
    ) -> Dict:
        try:
            result = self.session.request(
                method=method,
                params=params,
                url=self.base_url + relative_path,
                headers=headers,
                data=data,
                json=json,
            )
            result.raise_for_status()
        except requests.exceptions.RequestException as e:
            if e.response.status_code == 403:
                raise error.APIConnectionError("Connection forbidden. Please check the provided API key.")
            elif e.response.status_code in [400, 422]:
                raise e
            else:
                raise error.APIConnectionError("Error communicating with IAI") from e
        iai.logger.info(
            iai.logger.logfmt(
                "IAI API response",
                path=self.base_url,
                response_code=result.status_code,
            )
        )
        data = self._interpret_response_line(result)
        return data

    def _get_base_url(self) -> str:
        """
        This function returns the endpoint for API calls, which includes the
        version and other endpoint specifications.
        The method path should be appended to the base_url
        """
        if not iai.dev:
            base_url = "https://api.inverted.ai/v0/aws/m1"
        else:
            base_url = iai.dev_url
        # TODO: Add endpoint option and versioning to base_url
        return base_url

    def _handle_error_response(self, rbody, rcode, resp, rheaders, stream_error=False):
        try:
            error_data = resp["error"]
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
            )

        if "internal_message" in error_data:
            error_data["message"] += "\n\n" + error_data["internal_message"]

        iai.logger.info(
            iai.logger.logfmt(
                "IAI API error received",
                error_code=error_data.get("code"),
                error_type=error_data.get("type"),
                error_message=error_data.get("message"),
                error_param=error_data.get("param"),
                stream_error=stream_error,
            )
        )

        # Rate limits were previously coded as 400's with code 'rate_limit'
        if rcode == 429:
            return error.RateLimitError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode in [400, 404, 415]:
            return error.InvalidRequestError(
                error_data.get("message"),
                error_data.get("param"),
                error_data.get("code"),
                rbody,
                rcode,
                resp,
                rheaders,
            )
        elif rcode == 401:
            return error.AuthenticationError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 403:
            return error.PermissionError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 409:
            return error.TryAgain(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        else:
            return error.APIError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )

    def _interpret_response_line(self, result):
        rbody = result.content
        rcode = result.status_code
        rheaders = result.headers

        if rcode == 503:
            raise error.ServiceUnavailableError(
                "The server is overloaded or not ready yet.",
                rbody,
                rcode,
                headers=rheaders,
            )
        try:
            data = json.loads(rbody)
        except:
            raise error.APIError(
                f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders
            )
        if "error" in data or not 200 <= rcode < 300:
            raise self._handle_error_response(rbody, rcode, data, rheaders)

        return data


class APITokenAuth(AuthBase):
    def __init__(self, api_token):
        self.api_token = api_token

    def __call__(self, r):
        r.headers["x-api-key"] = self.api_token
        r.headers["api-key"] = self.api_token
        return r


def Jupyter_Render():
    import ipywidgets as widgets
    import matplotlib.pyplot as plt
    import numpy as np

    class Jupyter_Render(widgets.HBox):
        def __init__(self):
            super().__init__()
            output = widgets.Output()
            self.buffer = [np.zeros([128, 128, 3], dtype=np.uint8)]

            with output:
                self.fig, self.ax = plt.subplots(
                    constrained_layout=True, figsize=(5, 5)
                )
            self.im = self.ax.imshow(self.buffer[0])
            self.ax.set_axis_off()

            self.fig.canvas.toolbar_position = "bottom"

            self.max = 0
            # define widgets
            self.play = widgets.Play(
                value=0,
                min=0,
                max=self.max,
                step=1,
                description="Press play",
                disabled=False,
            )
            self.int_slider = widgets.IntSlider(
                value=0, min=0, max=self.max, step=1, description="Frame"
            )

            controls = widgets.HBox(
                [
                    self.play,
                    self.int_slider,
                ]
            )
            controls.layout = self._make_box_layout()
            widgets.jslink((self.play, "value"), (self.int_slider, "value"))
            output.layout = self._make_box_layout()

            self.int_slider.observe(self.update, "value")
            self.children = [controls, output]

        def update(self, change):
            self.im.set_data(self.buffer[self.int_slider.value])
            self.fig.canvas.draw()

        def add_frame(self, frame):
            self.buffer.append(frame)
            self.int_slider.max += 1
            self.play.max += 1
            self.int_slider.value = self.int_slider.max
            self.play.value = self.play.max

        def _make_box_layout(self):
            return widgets.Layout(
                border="solid 1px black",
                margin="0px 10px 10px 0px",
                padding="5px 5px 5px 5px",
            )

    return Jupyter_Render()


class IAILogger(logging.Logger):
    def __init__(
        self,
        name: str = "IAILogger",
        level: str = "WARNING",
        consoel: bool = True,
        log_file: bool = False,
    ) -> None:

        level = logging.getLevelName(level)
        log_level = level if type(level) == int else 30
        super().__init__(name, log_level)
        if consoel:
            consoel_handler = logging.StreamHandler()
            self.addHandler(consoel_handler)
        if log_file:
            file_handler = logging.FileHandler("iai.log")
            self.addHandler(file_handler)

    @staticmethod
    def logfmt(message, **params):
        props = dict(message=message, **params)

        def fmt(key, val):
            # Handle case where val is a bytes or bytesarray
            if hasattr(val, "decode"):
                val = val.decode("utf-8")
            # Check if val is already a string to avoid re-encoding into ascii.
            if not isinstance(val, str):
                val = str(val)
            if re.search(r"\s", val):
                val = repr(val)
            # key should already be a string
            if re.search(r"\s", key):
                key = repr(key)
            return f"{key}={val}"

        return " ".join([fmt(key, val) for key, val in sorted(props.items())])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invertedai', 'invertedai.simulators', 'invertedai.simulators.data']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.21.0,<0.22.0']

setup_kwargs = {
    'name': 'invertedai',
    'version': '0.0.3rc0',
    'description': 'Client SDK for InvertedAI',
    'long_description': '[pypi-badge]: https://badge.fury.io/py/invertedai.svg\n[pypi-link]: https://pypi.org/project/invertedai/\n\n\n[![Documentation Status](https://readthedocs.org/projects/inverted-ai/badge/?version=latest)](https://inverted-ai.readthedocs.io/en/latest/?badge=latest)\n[![PyPI][pypi-badge]][pypi-link]\n\n# InvertedAI\n## Overview\n<!-- start elevator-pitch -->\nInverted AI has trained cutting-edge realistic behavioral driving models that are human-like and close the SIM2Real. Our API provides access to these behavioral models and can be useful for several tasks in autonomous vehicle (AV) research and development.\n\n![](docs/images/top_camera.gif)\n<!-- end elevator-pitch -->\n\n# Get Started\n<!-- start quickstart -->\nIn this quickstart tutorial, you’ll run a simple sample AV simulation with Inverted AI Python API. Along the way, you’ll learn key concepts and techniques that are fundamental to using the API for other tasks. In particular, you will be familiar with two main Inverted AI models:\n\n- DRIVE\n- INITIALIZE\n\n## Quick Start\n- **Goolge Colab***: [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/inverted-ai/invertedai-drive/blob/develop/examples/Colab-Demo.ipynb)\n\n- **Locally**: Download the \n<a href="https://minhaskamal.github.io/DownGit/#/home?url=https://github.com/inverted-ai/invertedai/tree/master/examples" target="_blank">examples directory</a>, navigate to the unzipped directory in terminal and run \n\n```\npython3 -m venv .venv\nsource .venv/bin/activate\npip install --upgrade pip\npip install -r requirements.txt\n.venv/bin/jupyter notebook Drive-Demo.ipynb\n```\n\n## Installation\n\n[pypi-badge]: https://badge.fury.io/py/invertedai.svg\n[pypi-link]: https://pypi.org/project/invertedai/\n\nTo install use [![PyPI][pypi-badge]][pypi-link]:\n\n```bash\npip install invertedai\n```\n\n## Setting up\n\nImport the _invertedai_ package and set the API key with **add_apikey** method.\n\nRefer to the [product page](https://www.inverted.ai) to get your API key (or recharge for more tokens).\n\n```python\n\nimport invertedai as iai\niai.add_apikey("XXXXXXXXXXXXXX")\n```\n\n## Browse through available maps (location)\nInverted AI provides an assortment of diverse road configurations and geometries, including real-world locations and maps from simulators (CARLA, Huawei SMARTS, ...).\\\nTo search the catalog use **iai.available_maps** method by providing keywords\n```python\niai.available_maps("roundabout", "carla")\n```\n\nTo get information about a scene use **iai.get_map**.\n```python\niai.get_map**("CARLA:Town03:Roundabout")\n```\nThe scene information include the map in [Lanelet2](https://github.com/fzi-forschungszentrum-informatik/Lanelet2) format, map in JPEG format, maximum number of allowed (driveable) vehicles, latitude longitude coordinates (for real-world locations), id and list of traffic light and signs (if any exist in the map), etc.\n\n## INITIALIZE\nTo run the simulation, the map must be first populated with agents.\nInverted AI provides the **INITIALIZE**, a state-of-the-art model trained with real-life driving scenarios which can generate realistic positions for the initial state of the simulation.\\\nHaving realistic, complicated and diverse initial conditions are particularly crucial to observer interesting and informative interaction between the agents, i.e., the ego vehicle and NPCs (non-player characters).\n\nYou can use **INITIALIZE** in two modes:\n- _Initialize all agents_: generates initial conditions (position and speed) for all the agents including the ego vehicle\n```python\nresponse = iai.initialize(\n    location="CARLA:Town03:Roundabout",\n    agent_count=10,\n)\n```\n- _Initialize NPCs_: generates initial conditions (position and speed) only for the NPCs according to the provided state of the ego vehicle.\n```python\nresponse = iai.initialize(\n    location="CARLA:Town03:Roundabout",\n    agent_count=10,\n)\n```\n> _response_ is a dictionary of _states_, and _agent-attribute_  (_recurrent-states_ is also returned for compatibility with **drive**)\\\n> _response["states"]_ is a list of agent states, by default the first on the list is always the ego vehicle.\n\n## DRIVE\n**DRIVE** is Inverted AI\'s cutting-edge realistic driving model trained on millions of miles of traffic data.\nThis model can drive all the agents with only the current state of the environment, i.e., one step observations (which could be obtained from **INITIALIZE**) or with multiple past observations.\n```python\nresponse = iai.drive(\n    location="CARLA:Town03:Roundabout",\n    agent_attributes=response["attributes"],\n    states=response["states"],\n    recurrent_states=response["recurrent_states"],\n    traffic_states_id=response["traffic_states_id"],\n    steps=1,\n    get_birdviews=True,\n    get_infractions=True,\n)\n```\n>For convenience and to reduce data overhead, ***drive** also returns _recurrent-states_ which can be feedbacked to the model instead of providing all the past observations.\\\n>Furthermore, **drive** drive all the agents for $steps\\times \\frac{1}{FPS}$ where by default $FPS=10[frames/sec]$, should you require other time resolutions [contact us](mailto:info@inverted.ai).\n\n<!-- end quickstart -->\n',
    'author': 'Inverted AI',
    'author_email': 'info@inverted.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import os
from dotenv import load_dotenv
from invertedai.api_resources import drive, initialize, get_map, available_maps
from invertedai.utils import Jupyter_Render, IAILogger, Session

load_dotenv()
dev = os.environ.get("DEV", False)
if dev:
    dev_url = os.environ.get("DEV_URL", "http://localhost:8000")
log_level = os.environ.get("LOG_LEVEL", "WARNING")
log_console = os.environ.get("LOG_CONSOLE", 1)
log_file = os.environ.get("LOG_FILE", 0)
api_key = os.environ.get("API_KEY", "")

logger = IAILogger(level=log_level, consoel=bool(log_console), log_file=bool(log_file))

session = Session(api_key)
add_apikey = session.add_apikey
model_resources = {
    "initialize": ("get", "/initialize"),
    "drive": ("post", "/drive"),
    "get_map": ("get", "/map"),
    "available_maps": ("get", "/available_maps"),
}
try:
    from invertedai.simulators import CarlaEnv, CarlaSimulationConfig
except:
    logger.warning(
        "Cannot import CarlaEnv\n"
        + "Carla Python API is not installed\n"
        + "Ignore these warnings if you are not running Carla"
    )


__all__ = [
    "drive",
    "initialize",
    "get_map",
    "Jupyter_Render",
    "logger",
    "session",
    "add_apikey",
    "get_map",
    "available_maps",
]



from dataclasses import dataclass
from typing import List


@dataclass
class AgentStatesWithSample:
    x: List[List[List[float]]]  # BxAxT_obs
    y: List[List[List[float]]]  # BxAxT_obs
    psi: List[List[List[float]]]  # BxAxT_obs
    speed: List[List[List[float]]]  # BxAxT_obs


@dataclass
class DriveResponse:
    states: AgentStatesWithSample
    recurrent_states: List[List[List[List[List[float]]]]]  # BxAx2x64
    bird_view: List[int]
    collision = List[List[List[float]]]  # T_obsxBxA
    offroad = List[List[List[float]]]  # T_obsxBxA
    wrong_way = List[List[List[float]]]  # T_obsxBxA
